@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use base qw( Exporter );
 
-our $VERSION = '0.31_02';
+our $VERSION = '0.33_01';
 
 use AnyEvent::Stomper::Frame;
 use AnyEvent::Stomper::Error;
@@ -122,7 +122,9 @@ sub new {
 
   $self->_reset_internals;
   $self->{_input_queue}      = [];
-  $self->{_temp_queue}       = [];
+  $self->{_temp_input_queue} = [];
+  $self->{_write_queue}      = [];
+  $self->{_temp_write_queue} = [];
   $self->{_pending_receipts} = {};
   $self->{_subs}             = {};
 
@@ -245,6 +247,7 @@ sub _connect {
     on_rtimeout      => $self->_create_on_rtimeout,
     on_eof           => $self->_create_on_eof,
     on_error         => $self->_create_on_handle_error,
+    on_drain         => $self->_create_on_drain,
     on_read          => $self->_create_on_read,
   );
 
@@ -342,6 +345,23 @@ sub _create_on_handle_error {
   };
 }
 
+sub _create_on_drain {
+  my $self = shift;
+
+  weaken($self);
+
+  return sub {
+    return unless @{ $self->{_write_queue} };
+
+    $self->{_temp_write_queue} = $self->{_write_queue};
+    $self->{_write_queue} = [];
+
+    while ( my $cmd = shift @{ $self->{_temp_write_queue} } ) {
+      $cmd->{on_receipt}->();
+    }
+  };
+}
+
 sub _create_on_read {
   my $self = shift;
 
@@ -543,11 +563,9 @@ sub _push_write {
     }
   }
 
-  my $need_receipt;
   if ( exists $NEED_RECEIPT{ $cmd->{name} }
     || defined $cmd_headers->{receipt} )
   {
-    $need_receipt = 1;
     if ( $cmd->{name} eq 'CONNECT' ) {
       $self->{_pending_receipts}{CONNECTED} = $cmd;
     }
@@ -561,6 +579,9 @@ sub _push_write {
       $self->{_pending_receipts}{ $cmd_headers->{receipt} } = $cmd;
     }
   }
+  else {
+    push( @{ $self->{_write_queue} }, $cmd );
+  }
 
   my $body = $cmd->{body};
   unless ( defined $body ) {
@@ -581,10 +602,6 @@ sub _push_write {
 
   $self->{_handle}->push_write($frame_str);
 
-  unless ($need_receipt) {
-    AE::postpone { $cmd->{on_receipt}->() };
-  }
-
   return;
 }
 
@@ -680,10 +697,10 @@ sub _wtimeout {
 sub _process_input_queue {
   my $self = shift;
 
-  $self->{_temp_queue}  = $self->{_input_queue};
+  $self->{_temp_input_queue}  = $self->{_input_queue};
   $self->{_input_queue} = [];
 
-  while ( my $cmd = shift @{ $self->{_temp_queue} } ) {
+  while ( my $cmd = shift @{ $self->{_temp_input_queue} } ) {
     $self->_push_write($cmd);
   }
 
@@ -889,7 +906,9 @@ sub _abort {
   my %subs            = %{ $self->{_subs} };
 
   $self->{_input_queue}      = [];
-  $self->{_temp_queue}       = [];
+  $self->{_temp_input_queue} = [];
+  $self->{_write_queue}      = [];
+  $self->{_temp_write_queue} = [];
   $self->{_pending_receipts} = {};
   $self->{_subs}             = {};
 
@@ -930,7 +949,9 @@ sub _queued_commands {
 
   return (
     values %{ $self->{_pending_receipts} },
-    @{ $self->{_temp_queue} },
+    @{ $self->{_temp_write_queue} },
+    @{ $self->{_write_queue} },
+    @{ $self->{_temp_input_queue} },
     @{ $self->{_input_queue} },
   );
 }
@@ -1220,10 +1241,11 @@ For commands C<SUBSCRIBE>, C<UNSUBSCRIBE>, C<DISCONNECT> the client
 automatically adds C<receipt> header for internal usage.
 
 The command callback is called in one of two cases depending on the presence of
-the C<receipt> header. First case, when the command was successfully sent to
-the server. Second case, when the C<RECEIPT> frame will be received. If any
-error occurred during the command execution, the error object is passed to the
-callback in second argument. Error object is the instance of the class
+the C<receipt> header. First case, when the command was successfully written to
+the socket. Second case, when the C<RECEIPT> frame will be received. In first
+case C<on_receipt> callback can be called synchronously. If any error occurred
+during the command execution, the error object is passed to the callback in
+second argument. Error object is the instance of the class
 L<AnyEvent::Stomper::Error>.
 
 The command callback is optional. If it is not specified and any error
